# Copyright (C) 2015-2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Mikolaj Izdebski <mizdebsk@redhat.com>

from unittest import TestCase, skip
from time import sleep
from koschei.cache_manager import CacheManager

class RepoFactory(object):
    def create(self, repo_id, ignored):
        return "repo %s" % repo_id
    def destroy(self, repo_id, repo):
        assert repo == ("repo %s" % repo_id)
    def populate_cache(self):
        pass

class RepoFactory_Initial(RepoFactory):
    def populate_cache(self):
        return ((7541, 'repo 7541'), (689, 'repo 689'))

class RepoFactory_CreateException(RepoFactory):
    def create(self, repo_id, ignored):
        raise Exception("error downloading repo")

class RepoFactory_DestroyException(RepoFactory):
    def destroy(self, repo_id, repo):
        raise Exception("error removing repo from disk")

class RepoFactory_PopulateException(RepoFactory):
    def populate_cache(self):
        raise Exception("error loading repos from disk")

class SackFactory(object):
    def create(self, repo_id, repo):
        assert repo == ("repo %s" % repo_id)
        return "sack %s" % repo_id
    def destroy(self, repo_id, sack):
        assert sack == ("sack %s" % repo_id)
    def populate_cache(self):
        pass

class CacheManagerTest(TestCase):
    def setUp(self):
        self.mgr = CacheManager(2)
        self.mgr.add_bank(SackFactory(), 3, 2)
        self.mgr.add_bank(RepoFactory(), 5, 4)

    def test_termination(self):
        self.mgr.terminate()

    def test_single_item(self):
        self.mgr.prefetch(7541)
        self.mgr.acquire(7541)
        self.mgr.release(7541)
        self.mgr.terminate()

    def test_double_item(self):
        self.mgr.prefetch(7541)
        self.mgr.prefetch(689)
        self.mgr.acquire(7541)
        self.mgr.acquire(689)
        self.mgr.release(7541)
        self.mgr.release(689)
        self.mgr.terminate()

    def test_acquire_cached(self):
        self.mgr.prefetch(7541)
        self.mgr.acquire(7541)
        self.mgr.release(7541)
        self.mgr.prefetch(7541)
        self.mgr.acquire(7541)
        self.mgr.release(7541)
        self.mgr.terminate()

    def test_acquire_multiple(self):
        for i in xrange(1, 10):
            self.mgr.prefetch(i)
        sleep(1)
        for i in xrange(1, 10):
            self.mgr.acquire(i)
            sleep(0.05)
            self.mgr.release(i)
        self.mgr.terminate()

    # Regression test for issue #81
    def test_hit_miss_ordering(self):
        self.mgr = CacheManager(2)
        self.mgr.add_bank(SackFactory(), 1, 2)
        self.mgr.add_bank(RepoFactory(), 5, 4)
        self.mgr.prefetch(2)
        self.mgr.acquire(2)
        self.mgr.release(2)
        self.mgr.prefetch(1)
        self.mgr.prefetch(2)
        self.mgr.acquire(1)
        self.mgr.release(1)
        self.mgr.acquire(2)
        self.mgr.release(2)
        self.mgr.terminate()

    def test_acquire_twice(self):
        self.mgr.prefetch(7541)
        self.mgr.prefetch(7541)
        self.mgr.acquire(7541)
        self.mgr.release(7541)
        self.mgr.acquire(7541)
        self.mgr.release(7541)
        self.mgr.terminate()

    def test_initial_cache(self):
        self.mgr = CacheManager(1)
        self.mgr.add_bank(SackFactory(), 1, 2)
        self.mgr.add_bank(RepoFactory_Initial(), 1, 4)
        self.mgr.terminate()

    def test_create_exception(self):
        self.mgr = CacheManager(1)
        self.mgr.add_bank(SackFactory(), 1, 2)
        self.mgr.add_bank(RepoFactory_CreateException(), 1, 4)
        self.mgr.prefetch(7541)
        val = self.mgr.acquire(7541)
        assert not val
        self.mgr.release(7541)
        self.mgr.terminate()

    def test_destroy_exception(self):
        self.mgr = CacheManager(1)
        self.mgr.add_bank(SackFactory(), 1, 2)
        self.mgr.add_bank(RepoFactory_DestroyException(), 1, 4)
        self.mgr.prefetch(7541)
        self.mgr.acquire(7541)
        self.mgr.release(7541)
        self.mgr.terminate()

    def test_populate_exception(self):
        self.mgr = CacheManager(1)
        self.mgr.add_bank(SackFactory(), 1, 2)
        self.mgr.add_bank(RepoFactory_PopulateException(), 1, 4)
        self.mgr.prefetch(7541)
        self.mgr.acquire(7541)
        self.mgr.release(7541)
        self.mgr.terminate()
